#include "config.h"
#include "commands.h"
#include "support.h"
#include "io.h"
#include "sci.h"

#define MAX_PULSES 2048
 
unsigned long locaddr = 0;

unsigned short PLenRing[PRING_COUNT];
unsigned short PPatternRing[PRING_COUNT];
unsigned short PNextParseRing;

unsigned short Pulses[MAX_PULSES];
unsigned long PLongCount;
unsigned short PCurPattern;
unsigned short PCount, PIndex;

int CMDSetBaudRate (const char *buf, int fd)
{
unsigned long rate;
int stop, usedma;
int n;
  n = sscanf(buf, "%lu %d %d", &rate, &stop, &usedma);
  rate = rate / 100;
  if(n < 1) return -1;
  if(n < 2) stop = 0;
  if(n < 3) usedma = 1;
  fdprintf(fd, "rate:%lu, stop:%d" NEWLINE, rate, stop);
  InitSCI0(0, 0, stop, rate, usedma);
  return 0;  
}

int CMDPulseClear (const char *buf, int fd)
{
  PIndex = 0;
  PCount = 0;
  PIndex = 0; //ensure
  return 0;
}

void parse2ring(unsigned short pc)
{
unsigned short *p, *q;
unsigned short i, etime = 0;
  p = &PLenRing[PNextParseRing];
  q = &PPatternRing[PNextParseRing];
  for(i = PNextParseRing; i != pc; i++)
  {
	if(i >= PRING_COUNT)
	{
		p = PLenRing; q = PPatternRing; i = 0;
		if(pc == 0) break;
	}

	  if(Pulses[PIndex] < 0x8000u)
	  {
		*p = Pulses[PIndex];
		*q = PCurPattern = Pulses[PIndex + 1];
		PIndex += 2;
 	 }
 	 else
 	 {
		if(PLongCount <= 0)
		{
			PLongCount = (unsigned long)Pulses[PIndex + 1] * 0x8000u + Pulses[PIndex] - 0x8000u;
		}

		PLongCount--;

		if(PLongCount <= 0)
		{
			*p = Pulses[PIndex + 2];
			*q = PCurPattern = Pulses[PIndex + 3];
			PIndex += 4;	
		}
		else
		{
			*p = 0x7fff;
			*q = PCurPattern;
		}
  	}
  	if(PIndex >= PCount) PIndex = 0;
	etime += *p / 0x100;
	if(etime > 0x1000) break;
	p++;
	q++;	
  }
  PNextParseRing = i;
}

void PulseTryOn(void)
{
  if(DMAC0A.DTCR.BIT.DTE) return;
  DMAC0A.DTCR.BIT.DTE;
  DMAC0A.DTCR.BIT.DTE = 0;
  DMAC0B.DTCR.BIT.DTE;
  DMAC0B.DTCR.BIT.DTE = 0;

  ITU0.TIOR.BYTE = 0x00;
  ITU0.TCR.BYTE = 0x20;
  ITU0.TIER.BYTE = 0x00;
  ITU1.TIOR.BYTE = 0x00;
  ITU1.TCR.BYTE = 0x20;
  ITU1.TIER.BYTE = 0x00;
  ITU0.TCNT = 0xfffe;
  ITU0.GRA = 0xffff;
  ITU0.GRB = 0xffff;
  ITU0.TSR.BIT.IMFA;
  ITU0.TSR.BIT.IMFA = 0;
  ITU1.GRA = 0x1fff;
  ITU1.GRB = 0xffff;
  ITU1.TCNT = 0xffff;

  DMAC0A.MAR = PLenRing;
  DMAC0A.IOAR = (unsigned char)&ITU0.GRA;
  DMAC0A.ETCR = PRING_COUNT * 0x101;
  DMAC0A.DTCR.BIT.DTE;
  DMAC0A.DTCR.BYTE = 0x50;
  DMAC0B.MAR = PPatternRing;
  DMAC0B.IOAR = (unsigned char)&TPC.NDRB1.BYTE;
  DMAC0B.ETCR = PRING_COUNT * 0x101;
  DMAC0B.DTCR.BIT.DTE;
  DMAC0B.DTCR.BYTE = 0x50;

  PCurPattern = Pulses[PCount - 1];
  PIndex = 0; PLongCount = 0;
  PNextParseRing = 0;
  parse2ring(PRING_COUNT);

  DMAC0A.DTCR.BIT.DTE;
  DMAC0A.DTCR.BIT.DTE = 1;
  DMAC0B.DTCR.BIT.DTE;
  DMAC0B.DTCR.BIT.DTE = 1;

  ITU0.TIER.BYTE = 0x01;
  ITU1.TIER.BYTE = 0x01;

//  PA.DR.BYTE = PPattern[PCount - 1] % 0x100;
//  PB.DR.BYTE = PPattern[PCount - 1] / 0x100;
  PA.DDR = 0xff;
  PB.DDR = 0xff;
  TPC.NDERA.BYTE = 0xff;
  TPC.NDERB.BYTE = 0xff;
  TPC.TPCR.BYTE = 0x00;
  TPC.TPMR.BYTE = 0;
  TPC.NDRA1.BYTE = PPatternRing[0] % 0x100;
  TPC.NDRB1.BYTE = PPatternRing[0] / 0x100;
  STI();
  ITU.TSTR.BIT.STR0 = 1;
  ITU.TSTR.BIT.STR1 = 1;
}
int CMDPulseOn (const char *buf, int fd)
{
  fdprintf(fd, "Enabl");
  PulseTryOn();
  fdprintf(fd, "ed" NEWLINE);
  return 0;
}

int CMDPulseOff (const char *buf, int fd)
{
int n;
unsigned short d;
  n = sscanf(buf, "%x", &d);

  ITU.TSTR.BIT.STR1 = 0;
  ITU.TSTR.BIT.STR0 = 0;
  DMAC0A.DTCR.BIT.DTE;
  DMAC0B.DTCR.BIT.DTE;
  DMAC0A.DTCR.BIT.DTE = 0;
  DMAC0B.DTCR.BIT.DTE = 0;
  if(n < 1)
  {
  	PA.DDR = 0;
  	PB.DDR = 0;
  }
  else
  {
  	TPC.NDERA.BYTE = 0;
  	TPC.NDERB.BYTE = 0;
  	PA.DR.BYTE = d % 0x100;
	PB.DR.BYTE = d / 0x100;
  }
  return 0;
}

int CMDPulseLoad (const char *buf, int fd)
{
  int n, i;
  unsigned int cnt, inccnt;
  unsigned short sum, nsum;
  char c;

  n = sscanf(buf, "%x %x", &cnt, &inccnt);
  if(n < 1) return -1;
  if(n < 2) inccnt = cnt;

  fdprintf(fd, "%x %x" NEWLINE, cnt, inccnt);

  do {read(fd, &c, 1);} while(c != '>');
  
  for(i = 0; i < (int)cnt; i += inccnt)
  {
	if(read(fd, (char *) &Pulses[PCount], inccnt * 2) != (int)inccnt * 2)
    	{
		fdprintf(fd, "READ ERROR" NEWLINE);
      		return -1;
    	}
	sum = makesum((char *) &Pulses[PCount], inccnt * 2);
	read(fd, (char *) &nsum, 2);
	if(sum == nsum)
	{
		PCount += inccnt;
		PulseTryOn();
	}
  }
  do {read(fd, &c, 1);} while(c != '\n');
  fdprintf(fd, "%x" NEWLINE, PCount);
  if(PIndex >= PCount) PIndex = 0;
  return 0;
}

int CMDPulseAdd (const char *buf, int fd)
{
  int n;
  unsigned long ulen, dlen,  pattern;
  unsigned short llen;

// upper 32bit, least 32bit, pattern
  n = sscanf(buf, "%lu%lu%lx", &ulen, &dlen, &pattern);
  if(n < 3) return -1;

  if(dlen == 0) ulen--;
  dlen--;
  llen = dlen % 0x8000u;
  ulen = ulen * 0x20000u +  dlen / 0x8000u;


  fdprintf(fd, "%lu %lu %lx %lx %lx" NEWLINE, ulen, dlen, pattern, ulen, llen);
  
  switch(ulen)
    {
    case 0:
      Pulses[PCount] = llen;
      Pulses[PCount + 1] = pattern;
      PCount += 2;
      break;
    default:
      Pulses[PCount] = ulen % 0x8000u + 0x8000u;
      Pulses[PCount + 1] = ulen / 0x8000u;
      Pulses[PCount + 2] = llen + 0x8000u;
      Pulses[PCount + 3] = pattern;
      PCount += 4;
      break;      
    }
  return 0;
}


int CMDDump (const char *buf, int fd)
{
  unsigned long address, bytes, i, c;
  int n;

  n = sscanf(buf, "%lx%lx", &address, &bytes);
  if(n == 0) address = locaddr;
  if(n < 2) bytes = 0x100;
  fdprintf(fd, "START: %lX, STOP: %lX, BYTES: %lX" NEWLINE, 
	   address, (unsigned long)(address + bytes - 1), bytes);
  for(i = address, c = 0; c < bytes; i++, c++)
    {
      if(c % 16 == 0)
	fdprintf(fd, "%lX :", address);
      fdprintf(fd, " %.2X", (unsigned int) *((unsigned char *) address++));
      if(c % 16 == 15)
	fdprintf(fd, NEWLINE);
    }
  fdprintf(fd, NEWLINE);
  return 0;
}

int CMDGoLoc (const char *buf, int fd)
{
  unsigned long address;
  int n;

  n = sscanf(buf, "%lx", &address);
  if(n != 1) return -1;
  locaddr = address;
  return 0;
}

int CMDWriteByte (const char *buf, int fd)
{
  unsigned long address, data;
  int n;

  n = sscanf(buf, "%lx%lx", &address, &data);
  if(n != 2) return -1;
  *((unsigned char *) address) = (unsigned char) data;
  return 0;
}

int CMDWriteWord (const char *buf, int fd)
{
  unsigned long address, data;
  int n;

  n = sscanf(buf, "%lx%lx", &address, &data);
  if(n != 2) return -1;
  *((unsigned short *) address) = (unsigned short) data;
  return 0;
}

int CMDSleep (const char *buf, int fd)
{
  unsigned int ms;
  int n;

  n = sscanf(buf, "%u", &ms);
  if(n != 1) return -1;
  sleepinms(ms);
  fdprintf(fd, "OK" NEWLINE);
  return 0;
}


typedef void (*fp)(void);

int CMDLoad (const char *buf, int fd)
{
  int n;
  unsigned long address, bytes, sum, start;
  char c;

  start = 0;
  n = sscanf(buf, "%lx%lx%lx%lx", &address, &bytes, &sum, &start);
  if(n < 3) return -1;
  if(n < 4) start = 0;

  do {read(fd, &c, 1);} while(c != '>');

  if(read(fd, (char *) address, bytes) != (int)bytes)
    {
      fdprintf(fd, "COM ERROR" NEWLINE);
      return 0;
    }
  if(sum != makesum((void *)address, bytes))
    {
      fdprintf(fd, "CHECK SUM ERROR" NEWLINE);
      return 0;
    }
  if(start)
    {
      ((fp)address)(); 
    }
  fdprintf(fd, "CHECK SUM OK" NEWLINE);
  return 0;
}

int CMDRun (const char *buf, int fd)
{
  int n;
  unsigned long address;

  n = sscanf(buf, "%lx", &address);
  if(n != 1) return -1;
  ((fp)address)();
  return 0;
}
