#include "config.h"
#include "support.h"
#include "sci.h"
#include <string.h>

#define RDBufCount 128u
short RDBufIdx;
char RDBuf[RDBufCount];
#define WRBufCount 16u
char WRBuf[WRBufCount];
char UseDMA = 0;

#ifndef NOUSE_OPEN
int InitSCI0 (int parity, int odd, int stop, unsigned int baudratep100, int usedma)
{
  int dck, cks;
  unsigned int n;
  UseDMA = usedma;
  dck = 1;

  for(cks = 0; cks < 4; cks++)
    {
      n = (CLOCK_IN_100Hz / baudratep100 / 16 / dck  - 1) / 2;
      if(n < 256) break;
      dck *= 4;
    }

  SCI0.SCR.BIT.TE = 0;
  SCI0.SCR.BIT.RE = 0;

  //no SCK, no interrupts
  SCI0.SCR.BYTE = 0x00;
  //ensure
  SCI0.SCR.BYTE = 0x00;

  //asynchronous, 8bit, noparity, even
  SCI0.SMR.BYTE = cks;
  if(parity) SCI0.SMR.BIT.PE = 1;
  if(odd) SCI0.SMR.BIT.OE = 1;
  if(stop) SCI0.SMR.BIT.STOP = 1;

  SCI0.BRR = n;

  sleepinus(2 * n);

  if(UseDMA)
  {
  SCI0.SCR.BIT.RIE = 1;
  STI();
  }
  SCI0.SCR.BIT.TE = 1;
  FlushRDRFSCI0();
  return 0;
}

int CloseSCI0 (void)
{
  SCI0.SCR.BYTE = 0x00;
  DMAC1A.DTCR.BIT.DTE = 0;
  DMAC1B.DTCR.BIT.DTE = 0;
  return 0;
}

int FlushRDRFSCI0 (void)
{
  SCI0.SSR.BIT.ORER;
  SCI0.SSR.BIT.ORER = 0;
  SCI0.SSR.BIT.PER;
  SCI0.SSR.BIT.PER = 0;
  SCI0.SSR.BIT.FER;
  SCI0.SSR.BIT.FER = 0;
  SCI0.SSR.BIT.RDRF;
  SCI0.SSR.BIT.RDRF = 0;
  SCI0.SCR.BIT.RE = 0;
  if(UseDMA)
  {
  DMAC1A.MAR = RDBuf;
  DMAC1A.IOAR = (unsigned char)&SCI0.RDR;
  DMAC1A.ETCR = RDBufCount * 0x101u;
  DMAC1A.DTCR;
  DMAC1A.DTCR.BYTE = 0x15; //repeate mode
  RDBufIdx = 0;
  DMAC1A.DTCR.BIT.DTE;
  DMAC1A.DTCR.BIT.DTE = 1;
  DMAC1B.DTCR;
  DMAC1B.DTCR.BYTE = 0x04;
  }
  else
  {
     DMAC1A.DTCR.BIT.DTE = 0;
  }
  SCI0.SCR.BIT.RE = 1;
  return 0;
}
#endif
  
int WriteSCI0DMA (int cnt, const unsigned char *bytes)
{
unsigned short n;
unsigned char i;
  n = cnt;
  while(n != 0)
  {
  if(n > WRBufCount) i = WRBufCount; else i = n;
  n -= i;
  while(DMAC1B.DTCR.BIT.DTE);
  memcpy(WRBuf, bytes, i);

  SCI0.SCR.BIT.TIE = 0;
  DMAC1B.MAR = WRBuf;
  DMAC1B.IOAR = (unsigned char)&SCI0.TDR;
  DMAC1B.ETCR = i;
  DMAC1B.DTCR;
  DMAC1B.DTCR.BYTE = 0x04;
  DMAC1B.DTCR.BIT.DTE;
  DMAC1B.DTCR.BIT.DTE = 1;
  SCI0.SCR.BIT.TIE = 1;
  }      
  return cnt;
}
int WriteSCI0 (int cnt, const unsigned char *bytes)
{
  int n;
//  if(UseDMA) return WriteSCI0DMA(cnt, bytes);
  for(n = 0; n < cnt; n++)
    {
      while(!SCI0.SSR.BIT.TDRE);
      
      SCI0.TDR = *bytes++;
      SCI0.SSR.BIT.TDRE = 0;
    }
  while(!SCI0.SSR.BIT.TEND);
      
  return cnt;
}

int ReadSCI0DMA (int cnt, unsigned char *bytes)
{
  int n, i;
  for(n = 0; n < cnt; n++)
    {
	do {
	i = RDBufCount - DMAC1A.ETCR / 0x100;
	} while(i == RDBufIdx);
	*bytes = RDBuf[RDBufIdx];
	bytes++;
	RDBufIdx++;
	if(RDBufIdx == RDBufCount) RDBufIdx = 0;
    }
  return cnt;
}
int ReadSCI0 (int cnt, unsigned char *bytes)
{
  int n;
  if(UseDMA) return ReadSCI0DMA(cnt, bytes);
  for(n = 0; n < cnt; n++)
    {
      while(!SCI0.SSR.BIT.RDRF)
	{
	  if(SCI0.SSR.BIT.ORER || SCI0.SSR.BIT.FER || SCI0.SSR.BIT.PER)
	    {
	      SCI0.SSR.BIT.ORER = 0;
	      SCI0.SSR.BIT.PER = 0;
	      SCI0.SSR.BIT.FER = 0;
	      return n;
	    }
	}
      *bytes++ = SCI0.RDR;
      SCI0.SSR.BIT.RDRF = 0;
    }
  return cnt;
}
