//#include <fcntl.h>
//#include <unistd.h>
#include <string.h>
#include "config.h"
#include "commands.h"
#include "support.h"
#include "sci.h"
#include "io.h"


int AwaitCommand (int fd);
int DoCommand (char *buf, int fd);
int scifd;

#define MAX_BUF 80
char buf[MAX_BUF];
int buflen;

extern void start(void);

short detectstack __attribute__ ((section(".detects")));

typedef void (*fp)(void);

#pragma interrupt
void Unhandled(void)
{
  fdprintf(scifd, "Unhandled Vector" NEWLINE);
  return;
}
const fp VectorTable[] __attribute__ ((section(".vectors"))) = 
{
  start, Unhandled, Unhandled, Unhandled, //0x00
  Unhandled, Unhandled, Unhandled, Unhandled, //0x10
  Unhandled, Unhandled, Unhandled, Unhandled, //0x20
  Unhandled, Unhandled, Unhandled, Unhandled, //0x30
  Unhandled, Unhandled, Unhandled, Unhandled, //0x40
  Unhandled, Unhandled, Unhandled, Unhandled, //0x50
  Unhandled, Unhandled, Unhandled, Unhandled,//0x60
  ITU1IMA, Unhandled, Unhandled, Unhandled, //0x70
  Unhandled, Unhandled, Unhandled, Unhandled, //0x80
  Unhandled, Unhandled, Unhandled, Unhandled, //0x90
  Unhandled, Unhandled, Unhandled, Unhandled, //0xa0
  Unhandled, Unhandled, Unhandled, Unhandled, //0xb0
  Unhandled, Unhandled, Unhandled, Unhandled, //0xc0
  SCI0ERI, Unhandled, SCI0TXI, Unhandled, //0xd0
  Unhandled, Unhandled, Unhandled, Unhandled, //0xe0
  Unhandled, Unhandled, Unhandled, Unhandled //0xf0
};


#pragma interrupt
void SCI0TXI(void)
{
  SCI0.SSR.BIT.TDRE;
  SCI0.SSR.BIT.TDRE = 0;
}
#pragma interrupt
void SCI0ERI(void)
{
  SCI0.SSR.BIT.ORER;
  SCI0.SSR.BIT.FER;
  SCI0.SSR.BIT.PER;

  SCI0.SSR.BIT.ORER = 0;
  SCI0.SSR.BIT.PER = 0;
  SCI0.SSR.BIT.FER = 0;
}


#pragma interrupt
void ITU1IMA(void)
{
  parse2ring(PRING_COUNT - DMAC0A.ETCR / 0x100);
  ITU1.TSR.BIT.IMFA;
  ITU1.TSR.BIT.IMFA = 0;
  return;
}

const struct
{
  const char *cmd;
  int (*func)(const char *buf, int fd);
} cmds[] =
  {
    {"d", CMDDump},
    {"dump", CMDDump},
    {"g", CMDGoLoc},
    {"goto", CMDGoLoc},
    {"wb", CMDWriteByte},
    {"writebyte", CMDWriteByte},
    {"ww", CMDWriteWord},
    {"writeword", CMDWriteWord},
    {"sleep", CMDSleep},
    {"load", CMDLoad},
    {"run", CMDRun},
    {"setbaudrate", CMDSetBaudRate},
    {"pon", CMDPulseOn},
    {"poff", CMDPulseOff},
    {"pclear", CMDPulseClear},
    {"padd", CMDPulseAdd},
    {"pload", CMDPulseLoad},
    //    {"pulsequery", CMDPulseQuery},
    //{"pulsecount", CMDPulseCount},
    {NULL, NULL}
  };

int main(void)
{
  int sci;

  sleepinms(50);
  PA.DDR = 0x40;
  PA.DR.BIT.B6 = 0; //FWE = 0;

  InitSCI0(0, 0, 0, DEF_BAUDRATE/100, 1);
  scifd = sci = open(DEV_SCI0, 0);
  sleepinms(200);
  WriteSCI0(2, "H8");
  fdprintf(sci, "/300H monitor" NEWLINE);

  locaddr = 0;

  detectstack = 0x3782;
  for(;;)
    {
      AwaitCommand(sci);
      if(detectstack != 0x3782) fdprintf(sci, "Stack violation" NEWLINE);
    }

  return 0;
}

int AwaitCommand (int fd)
{
  char s;
 char echo_on = 1;

  buf[0] = 0;
  buflen = 0;

  for(;;)
    {
      read(fd, &s, 1);
      if(s == '$') {echo_on = 0; continue;}
      if((s == '\n') || (s == '\r'))
	{
	  if(echo_on) fdprintf(fd, NEWLINE);
       	  buf[buflen] = 0;
	  if(buflen == 0) return -1;
	  return DoCommand(buf, fd);
	}

      if(echo_on) write(fd, &s, 1);
      buf[buflen] = s;
      buflen++;
      if(buflen >= MAX_BUF) return -1;
    }
}

int DoCommand (char *buf, int fd)
{
  int i, ret, len;

  for(i = 0;; i++)
    {
      if(!cmds[i].func)
	{
	  fdprintf(fd, "UNKNOWN COMMAND: %s" NEWLINE, buf);
	  return -1;
	}
      len = strlen(cmds[i].cmd);
      if((!strncmp(buf, cmds[i].cmd, len)) &&
	 ((buf[len] == ' ') || (buf[len] == '\0')))
	{
	  ret = (cmds[i].func)(buf + len, fd);
	  if(ret != 0)
	    fdprintf(fd, "SYNTAX ERROR:%d" NEWLINE, ret);
	  return ret;
	}
    }
	  
}
