#include "config.h"
#include "support.h"
#include "sci.h"

#ifndef NOUSE_OPEN
int InitSCI0 (int parity, int odd, int stop, int baudrate)
{
  int dck, cks, n;

  dck = 1;
  for(cks = 0; cks < 4; cks++)
    {
      n = CLOCK_IN_Hz / baudrate / 32 / dck  - 1;
      if(n < 256) break;
      dck *= 4;
    }

  //no SCK, no interrupts
  SCI0.SCR.BYTE = 0x00;
  //ensure
  SCI0.SCR.BYTE = 0x00;

  //asynchronous, 8bit, noparity, even
  SCI0.SMR.BYTE = cks;
  if(parity) SCI0.SMR.BIT.PE = 1;
  if(odd) SCI0.SMR.BIT.OE = 1;
  if(stop) SCI0.SMR.BIT.STOP = 1;

  SCI0.BRR = n;

  sleepinus(1500000 / baudrate);
  
  SCI0.SCR.BIT.TE = 1;
  SCI0.SCR.BIT.RE = 1;

  return 0;
}

int CloseSCI0 (void)
{
  SCI0.SCR.BYTE = 0x00;
  return 0;
}

int FlushRDRFSCI0 (void)
{
  SCI0.SSR.BIT.ORER;
  SCI0.SSR.BIT.ORER = 0;
  SCI0.SSR.BIT.PER;
  SCI0.SSR.BIT.PER = 0;
  SCI0.SSR.BIT.FER;
  SCI0.SSR.BIT.FER = 0;
  SCI0.SSR.BIT.RDRF;
  SCI0.SSR.BIT.RDRF = 0;
  return 0;
}
#endif

  
int WriteSCI0 (int cnt, unsigned char *bytes)
{
  int n;

  for(n = 0; n < cnt; n++)
    {
      while(!SCI0.SSR.BIT.TDRE);
      
      SCI0.TDR = *bytes++;
      SCI0.SSR.BIT.TDRE = 0;
    }
  while(!SCI0.SSR.BIT.TEND);
      
  return cnt;
}

int ReadSCI0 (int cnt, unsigned char *bytes)
{
  int n;

  for(n = 0; n < cnt; n++)
    {
      while(!SCI0.SSR.BIT.RDRF)
	{
	  if(SCI0.SSR.BIT.ORER || SCI0.SSR.BIT.FER || SCI0.SSR.BIT.PER)
	    {
	      SCI0.SSR.BIT.ORER = 0;
	      SCI0.SSR.BIT.PER = 0;
	      SCI0.SSR.BIT.FER = 0;
	      return n;
	    }
	}
      *bytes++ = SCI0.RDR;
      SCI0.SSR.BIT.RDRF = 0;
    }
  return cnt;
}
