#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <fcntl.h>

#ifdef USE_FP
#include <stdio.h>
#endif

#include "sio.h"

typedef unsigned char BOOL;

int nrvsscanf (const char *str, const char *fmt, va_list ap);
int _nrvprintf(int (*func)(void *, const char *, int), void *farg, const char *fmt, va_list ap);
int nrwrite(void *fd, const char *buf, int cnt);

int nrsscanf (const char *str, const char *fmt, ...)
{
  int ret;

  va_list ap;
  va_start(ap, fmt);
  ret = nrvsscanf( str, fmt, ap);
  va_end(ap);
  return ret;
}

int nrvsscanf (const char *str, const char *fmt, va_list ap)
{
  BOOL isfailed = 0;
  BOOL nolet, length, gonext, longval;
  short sign;
  short cnt = 0;
  void *val;
  long sinteger;
  unsigned long uinteger;

  for(; *fmt; fmt++)
    {
      switch(*fmt)
	{
	case ' ':
	  while(isspace(*str))  str++;
	  break;
	case '%':
	  while(isspace(*str))  str++;
	  if(!*str) isfailed = 1;
	  if(!*++fmt) break;
	  nolet = 0;
	  length = -1;
	  gonext = 0;
	  longval = 0;
	  for(; *fmt && !gonext;)
	    {
	      switch(*fmt)
		{
		case '*':
		  nolet = 1;
		  fmt++;
		  break;
		case 'l':
		  longval = 1;
		  fmt++;
		  break;
		default:
		  if(isdigit(*fmt))
		    {
		      length = 0;
		      for(; isdigit(*fmt); fmt++)
			{
			  length = 10 * length + (int)(*fmt - '0');
			}
		    }
		  else
		    {
		      gonext = 1;
		    }
		  break;
		}
	    }
	  switch(*fmt)
	    {
	    case '\0':
	      break;
	    case 's':
	      if(!nolet) val = va_arg(ap, char *);
	      if(!isfailed)
		{
		  isfailed = 1;
		  for(; !isspace(*str) && *str && length; str++, length--)
		    {
		      isfailed = 0;
		      if(!nolet) *(char *)val++ = *str;
		    }
		  if(!nolet && !isfailed)
		    {
		      *(char *)val = '\0';
		      cnt++;
		    }
		}
	      break;
	    case 'd':
	      if(!nolet)
		{
		  if(longval)
		    {
		      val = va_arg(ap, long *);
		      *(long *)val = 0;
		    }
		  else
		    {
		      val = va_arg(ap, int *);
		      *(int *)val = 0;
		    }
		}
	      sinteger = 0;
	      if(!isfailed)
		{
		  sign = 1;
		  if((*str == '-') || (*str == '+'))
		    {
		      if(*str == '-') sign = -1;
		      str++;
		    }
		  isfailed = 1;
		  for(; isdigit(*str) && length; str++, length--)
		    {
		      isfailed = 0;
		      sinteger = 10 * sinteger + (long)(*str - '0');
		    }
		  if(!nolet && !isfailed)
		    {
		      sinteger *= sign;
		      if(longval)
			*(long *)val = sinteger;
		      else
			*(int *)val = sinteger;
		      cnt++;
		    }
		}
	      break;
	    case 'u':
	      if(!nolet)
		{
		  if(longval)
		    {
		      val = va_arg(ap, unsigned long *);
		      *(unsigned long *)val = 0;
		    }
		  else
		    {
		      val = va_arg(ap, unsigned int *);
		      *(unsigned int *)val = 0;
		    }
		}
	      uinteger = 0;
	      if(!isfailed)
		{
		  isfailed = 1;
		  for(; isdigit(*str) && length; str++, length--)
		    {
		      isfailed = 0;
		      isfailed = 0;
		      uinteger = 10 * uinteger + (long)(*str - '0');
		    }
		  if(!nolet && !isfailed)
		    {
		      if(longval)
			*(unsigned long *)val = uinteger;
		      else
			*(unsigned int *)val = uinteger;
		      cnt++;
		    }
		}
	      break;
	    case 'X':
	    case 'x':
	      if(!nolet)
		{
		  if(longval)
		    {
		      val = va_arg(ap, unsigned long *);
		      *(unsigned long *)val = 0;
		    }
		  else
		    {
		      val = va_arg(ap, unsigned int *);
		      *(unsigned int *)val = 0;
		    }
		}
	      uinteger = 0;
	      if(!isfailed)
		{
		  isfailed = 1;
		  for(; isxdigit(*str) && length; str++, length--)
		    {
		      isfailed = 0;
		      if(!nolet)
			{
			  if(isdigit(*str))
			    uinteger =
			      16 * uinteger + (int)(*str - '0');
			  else
			    uinteger = 
			      16 * uinteger + 10 + (int)(tolower(*str) - 'a');
			}
		    }
		  if(!nolet && !isfailed)
		    {
		      if(longval)
			*(unsigned long *)val = uinteger;
		      else
			*(unsigned int *)val = uinteger;
		      cnt++;
		    }
		}
	      break;	      
	    default:
	      if(*fmt != *str++)
		{
		  isfailed = 1;
		  break;
		}
	    }
	  //	  while(isspace(*str))  str++;
	  break;
	default:
	  if(*fmt != *str++)
	    {
	      isfailed = 1;
	      break;
	    }
	}
    }
  return cnt;
}

int nrwrite(void *fd, const char *buf, int cnt)
{
  write(*(int *)fd, buf, cnt);
  return cnt;
}

int nrfdprintf(int fd, const char *fmt, ...)
{
  int ret;

  va_list ap;
  va_start(ap, fmt);
  ret = 
#ifdef SIM_DEBUG
    vprintf(fmt, ap);
#else
    _nrvprintf(nrwrite, &fd, fmt, ap);
#endif
  va_end(ap);
  return ret;
}

#ifdef USE_FP
int nrfputc(void *fp, const char *buf, int cnt)
{
  int i;

  for(i = 0; i < cnt; i++)
    {
      fputc(*buf++, (FILE *)fp);
    }
  return cnt;
}

int nrprintf(const char *fmt, ...)
{
  int ret;
  va_list ap;

  va_start(ap, fmt);
  ret = _nrvprintf(nrfputc, stdout, fmt, ap);
  va_end(ap);
  return ret;
}

int nrfprintf(FILE *fp, const char *fmt, ...)
{
  int ret;
  va_list ap;

  va_start(ap, fmt);
  ret = _nrvprintf(nrfputc, fp, fmt, ap);
  va_end(ap);
  return ret;
}
#endif

int _nrvprintf(int (*func)(void *, const char *, int), void *farg, const char *fmt, va_list ap)
{
  BOOL leftsided, dispsign, zeropadding, spaceforplus;
  BOOL issetacc, gonext, longval;
  short min, i, j, ds, accuracy;
  long dval, k;
  unsigned long uval, uval2;
  char *sval;
  char cspace;
  unsigned char alpha;
  char buf[16];
  short cnt = 0;

  const char space = ' ';
  const char minus = '-';
  const char plus = '+';

  for(; *fmt; fmt++)
    {
      i = 0;
      while((*(fmt + i) != '%') && *(fmt + i)) i++;
      if(i)
	{
	  (*func)(farg, fmt, i);
	  cnt += i;
	  fmt += i;
	}
      if(!*fmt) break;
      fmt++;
      leftsided = 0;
      dispsign = 0;
      zeropadding = 0;
      spaceforplus = 0;
      gonext = 0;
      longval = 0;
      for(; *fmt; fmt++)
	{
	  switch(*fmt)
	    {
	    case '-':
	      leftsided = 1;
	      break;
	    case '+':
	      dispsign = 1;
	      break;
	    case '0':
	      zeropadding = 1;
	      break;
	    case ' ':
	      spaceforplus = 1;
	      break;
	    case 'l':
	      longval = 1;
	      break;
	    default:
	      gonext = 1;
	      break;
	    }
	  if(gonext) break;
	}
      if(dispsign) spaceforplus = 0;
      if(leftsided) zeropadding = 0;

      min = 0;
      if(!*fmt) break;
      if(*fmt == '*')
	{
	  dval = longval ? va_arg(ap, long) : va_arg(ap, int);
	  min = dval;
	  if(dval < 0)
	    {
	      min = -dval;
	      leftsided = 1;
	    }
	  fmt++;
	}
      else
	{
	  for(;isdigit(*fmt); fmt++)
	    min = 10 * min + (long)(*fmt - '0');
	}
 
      issetacc = 0;
      if(*fmt == '.')
	{
	  issetacc = 1;
	  accuracy = 0;
	  fmt++;
	  if(*fmt == '*')
	    {
	      dval = longval ? va_arg(ap, long) : va_arg(ap, int);
	      if(dval > 0) accuracy = dval;
	      fmt++;
	    }
	  else
	    {
	      for(;isdigit(*fmt); fmt++)
		accuracy = 10 * accuracy + (long)(*fmt - '0');
	    }
	}
      if(!issetacc) accuracy = 1;

      if(!*fmt) break;
      alpha = 'a';
      cspace = ' ';
      if(zeropadding && !issetacc) cspace = '0';

      switch(*fmt)
	{
	case 's':
	  sval = va_arg(ap, char *);
	  if(issetacc)
	    {
	      i = (strlen(sval) > accuracy) ? accuracy : strlen(sval);
	    }
	  else
	    {
	      i = strlen(sval);
	    }
	  j = (i > min) ? 0 : min - i;
	  cnt += i + j;
	  if(leftsided)
	    {
	      (*func)(farg, sval, i);
	      for(; j > 0; j--) (*func)(farg, &space, 1);
	    }
	  else
	    {
	      for(; j > 0; j--) (*func)(farg, &space, 1);
	      (*func)(farg, sval + j, i);
	    }
	  break;
	case 'i':
	case 'd':
	  dval = longval ? va_arg(ap, long) : va_arg(ap, int);

	  ds = 0;
	  k = dval;
	  if(k < 0) k = -k;
	  while(k || ds < accuracy)
	    {
	      buf[ds] = (unsigned char)(k % 10) + '0';
	      k /= 10;
	      ds++;
	    }
	  i = ds;
	  if((dval < 0) || dispsign || spaceforplus) i++;
	  j = (i > min) ? 0 : min - i;
	  cnt += i + j;

	  if(!leftsided && !zeropadding)
	      for(; j > 0; j--) (*func)(farg, &cspace, 1);
	  if(dval < 0) (*func)(farg, &minus, 1);
	  if((dval >= 0) && dispsign) (*func)(farg, &plus, 1);
	  if((dval >= 0) && spaceforplus) (*func)(farg, &space, 1);
	  if(!leftsided && zeropadding)
	      for(; j > 0; j--) (*func)(farg, &cspace, 1);
	  for(; ds > 0; ds--)
	      (*func)(farg, &buf[ds - 1], 1);
	  if(leftsided)
	      for(; j > 0; j--) (*func)(farg, &cspace, 1);
	  break;
	case 'u':
	  uval = longval ? va_arg(ap, unsigned long) : va_arg(ap, unsigned int);

	  ds = 0;
	  uval2 = uval;
	  while(uval2 || ds < accuracy)
	    {
	      buf[ds] = (unsigned char)(uval2 % 10) + '0';
	      uval2 /= 10;
	      ds++;
	    }
	  i = ds;
	  j = (i > min) ? 0 : min - i;
	  cnt += i + j;

	  if(!leftsided)
	      for(; j > 0; j--) (*func)(farg, &cspace, 1);
	  for(; ds > 0; ds--)
	      (*func)(farg, &buf[ds - 1], 1);
	  if(leftsided)
	      for(; j > 0; j--) (*func)(farg, &cspace, 1);
	  break;
	case 'X':
	  alpha = 'A';
	case 'x':
	  uval = longval ? va_arg(ap, unsigned long) : va_arg(ap, unsigned int);

	  ds = 0;
	  uval2 = uval;
	  while(uval2 || ds < accuracy)
	    {
	      i = uval2 % 16;
	      buf[ds] = (unsigned char)i + ((i > 9) ? (alpha - 10) : '0');
	      uval2 /= 16;
	      ds++;
	    }
	  i = ds;
	  j = (i > min) ? 0 : min - i;
	  cnt += i + j;

	  if(!leftsided)
	      for(; j > 0; j--) (*func)(farg, &cspace, 1);
	  for(; ds > 0; ds--)
	      (*func)(farg, &buf[ds - 1], 1);
	  if(leftsided)
	      for(; j > 0; j--) (*func)(farg, &cspace, 1);
	  break;
	default:
	  (*func)(farg, fmt, 1);	  
	  cnt++;
	}
    }

  return cnt;
}
