#include "config.h"
#include "support.h"

//#define tP 10000
//#define tPMAX 200000
//#define tE 100000
//#define tEMAX 1200000
//#define NWEC 100
#define tSSWE 1
#define tSPSU 50
#define tSP30 30
#define tSP200 200
#define tSP10 10
#define tCP 5
#define tCPSU 5
#define tSPV 4
#define tSPVR 2
#define tCPV 2
#define tCSWE 100
#define NWMAX 1000
#define tSESU 100
#define tSE 10 //in ms
#define tSEMAX 20 // in ms
#define tCE 10
#define tCESU 10
#define tSEV 20
#define tSEVR 2
#define tCEV 4
#define tCSWE 100
#define NEMAX 120

#define SEQSIZE 128

struct _FLMCR {                          /*  Bit  Access */
  unsigned char FLER:1;      /*    FLER      */
  unsigned char SWE:1;      /*    SWE2      */
  unsigned char ESU:1;      /*    ESU2      */
  unsigned char PSU:1;      /*    PSU2      */
  unsigned char EV :1;      /*    EV2       */
  unsigned char PV :1;      /*    PV2       */
  unsigned char E  :1;      /*    E2        */
  unsigned char P  :1;      /*    P2        */
};                /*              */

typedef volatile struct _FLMCR FLMCR;
typedef volatile unsigned char EBR;

#define blkaddr(blk) (void *)((blk < 9) ? ((unsigned int)blk << 12) : ((unsigned int)(blk - 8) << 16))  

int WritePulse(FLMCR *flmcr, short writecnt);
int PGM(unsigned char blk, void *start, unsigned short sum);
int Erase(unsigned char blk);
int Read(void * address, unsigned short cnt);

#define StartPGM() (PA.DR.BIT.B6 = 1, PA.DDR = 0x40, 0)
#define FinishPGM()  (PA.DR.BIT.B6 = 0, 0)


#define TCSR (*(volatile short *)0xFFFFA8)
#define WRITE_TCSR(byte) (TCSR = 0xa500 + (unsigned char)(byte))
#define WRITE_TCNT(byte) (TCSR = 0x5a00 + (unsigned char)(byte))
#define RSTCSR (*(volatile short *)0xFFFFAA)
#define WRITE_RSTCSR(byte) (RSTCSR = 0xa500)

#define WDTSetupInms()   WRITE_TCSR(0x76)
#define WDTEnableInms(ms) WRITE_TCNT(0xff ^ (unsigned char)((ms) * (CLOCK_IN_Hz / 1000) / 2048))
#define WDTSetupInus() WRITE_TCSR(0x72)
#define WDTEnableInus(us) WRITE_TCNT(0xff ^ (unsigned char)((us) * (CLOCK_IN_Hz / 1000000) / 64))

#define WDTDisable() WRITE_TCSR(0)

#define Halt() {for(;;) asm("sleep");}
#define Reboot() {WDTSetupInus(); WDTEnableInus(1); for(;;);}

unsigned char buf[SEQSIZE * 3];

unsigned char fdgetc()
{
  unsigned char c;
  ttyread(1, &c);
  return c;
}
unsigned short fdget16()
{
  unsigned char c[2];
  unsigned short i;
  ttyread(2, c);  //MSB first
  i = (c[0] << 8) + c[1];
  return i;
}
void *fdget24()
{
  unsigned char c[3];
  unsigned int i;
  ttyread(3, c); //MSB first
  i = ((unsigned int)c[0] << 16) + (c[1] << 8) + c[2];
  return (void *)i;
}

void main(void)
{
  char c;
  unsigned char ret, blk;
  unsigned short cnt, sum;
  void * addr;
  int i;

  for(;;)
    {
      c = ':';
      ttywrite(1, &c);
      c = fdgetc();
      switch(c)
	{
	case 'c':
	  sum = fdget16();
	  for(i = 0; i < sum; i++) sleepinus(100);
	  ret = sum >> 8;
	  break;
	case 's':
	  ret = StartPGM();
	  break;
	case 'f':
	  ret = FinishPGM();
	  break;
	case 'p':
	  blk = fdgetc();
	  addr = (void *)fdget24();
	  sum = fdget16();
	  ret = PGM(blk, addr, sum);
	  break;
	case 'e':
	  blk = fdgetc();
	  ret = Erase(blk);
	  break;
	case 'r':
	  addr = (void *)fdget24();
	  cnt = fdget16();
	  ret = Read(addr, cnt);
	  break;
	case 'h':
	default:
	  //	  Halt();
	  Reboot();
	  break;
	}
      ttywrite(1, &ret);
    }
}

int Read(void * address, unsigned short cnt)
{
  unsigned char c;

  c = makesum(address, cnt) % 0x100; // LSB only
  ttywrite(1, &c);

  ttywrite(cnt, address);
  return 0;
}

int Erase(unsigned char blk)
{
  unsigned char cnt;
  char m;
  unsigned char n, verifybyte;
  volatile unsigned char *addr;

  unsigned short verifyword;

  unsigned char ebrbyte;

  FLMCR *flmcr;

  flmcr = (blk < 8) ? (void *)&FLASH.FLMCR1 : (void *)&FLASH.FLMCR2;

  for(n = 1; n <= NEMAX; n++)
    {
      m = 0;

      flmcr->SWE = 1;
      sleepinus(tSSWE);

      ebrbyte = 1 << (blk & 0x0f);
      if(blk < 8)
	FLASH.EBR1.BYTE = ebrbyte;
      else
	FLASH.EBR2.BYTE = ebrbyte;

      WDTSetupInms();
      WDTEnableInms(tSEMAX + (tSESU +  tCE + tCESU) / 1000);
      flmcr->ESU = 1;
      sleepinus(tSESU);
      flmcr->E = 1;
      sleepinms(tSE);
      flmcr->E = 0;
      sleepinus(tCE);
      flmcr->ESU = 0;
      sleepinus(tCESU);
      WDTDisable();

      flmcr->EV = 1;
      sleepinus(tSEV);

      addr = blkaddr(blk);

      for(cnt = SEQSIZE; cnt > 0; cnt--)
	{
	  *addr = 0xff;
	  sleepinus(tSEVR);

	  verifyword = *(unsigned short *)((int)addr & ~1);
	  addr++;

	  if((cnt & 1) == 0)
	    verifybyte = verifyword >> 8;
	  else
	    verifybyte = verifyword & 0xff;
	  
	  if(verifybyte != 0xff)
	    {
	      m = -1;
	      break;
	    }
	}
      flmcr->EV = 0;
      sleepinus(tCEV);
      if(m == 0)
	{
	  /* succeeded to erase */
	  break;
	}
    }    

  flmcr->SWE = 0;
  sleepinus(tCSWE);
  return m;
}

int PGM(unsigned char blk, void *start, unsigned short sum)
{
  short writecnt;
  
  volatile unsigned char *addr;
  unsigned char *pbuf;
  unsigned short verifyword;
  unsigned char verifybyte;
  unsigned char cnt;
  char m;

#define PREWRITE SEQSIZE
#define PAPPENDWRITE SEQSIZE * 2

  FLMCR *flmcr;

  flmcr = (blk < 8) ? (void *)&FLASH.FLMCR1 : (void *)&FLASH.FLMCR2;

  pbuf = buf;

  ttyread(SEQSIZE, pbuf);

  flmcr->SWE = 1;
  sleepinus(tSSWE);
  
  for(cnt = SEQSIZE; cnt > 0; cnt--)
    {
      *(pbuf + PREWRITE) = *pbuf;
      pbuf++;
    }

  if(makesum(buf + PREWRITE, SEQSIZE) != sum) return 'm';
  
  for(writecnt = 1; writecnt < NWMAX; writecnt++)
    {
      m = 0;

      addr = start;
      pbuf = buf;

      for(cnt = SEQSIZE; cnt > 0; cnt--)
	{
	  *addr++ = *(pbuf + PREWRITE);
	  pbuf++;
	}


      WritePulse(flmcr, writecnt);

      flmcr->PV = 1;
      sleepinus(tSPV);

      addr = start;
      pbuf = buf;

      for(cnt = SEQSIZE; cnt > 0; cnt--)
	{
	  *addr = 0xff;
	  sleepinus(tSPVR);
	  verifyword = *(unsigned short *)((int)addr & ~1);
	  addr++;

	  if((cnt & 1) == 0)
	    verifybyte = verifyword >> 8;
	  else
	    verifybyte = verifyword & 0xff;

	  if(verifybyte != *pbuf)  m = -1;
	  
	  //	  if(writecnt <= 6)
	  // {
	  *(pbuf + PAPPENDWRITE) = verifybyte | *(pbuf + PREWRITE);
	  //  }
	  
	  *(pbuf + PREWRITE) = *pbuf | ~verifybyte;
	  pbuf++;
	}

      flmcr->PV = 0;
      sleepinus(tCPV);

      if(writecnt < 6)
	{
	  addr = start;
	  pbuf = buf;
	  for(cnt = SEQSIZE; cnt > 0; cnt--)
	    {
	      *addr++ = *(pbuf + PAPPENDWRITE);
	      pbuf++;
	      WritePulse(flmcr, 0);
	    }
	}

      if(m == 0)
	{
	  /* succeeded to write */
	  break;
	}
    }

  flmcr->SWE = 0;
  sleepinus(tCSWE);
  return m;
}

int
WritePulse(FLMCR *flmcr, short writecnt)
{
  unsigned short steptsp;

  steptsp = STEP_SLEEPINUS(tSP10);
  WDTSetupInus();
  WDTEnableInus((tSP10 + tSPSU + tCP + tCPSU) * 3 / 2);
  if(writecnt > 0)
    {
      steptsp = STEP_SLEEPINUS(tSP200);
      WDTEnableInus((tSP200 + tSPSU + tCP + tCPSU) * 3 / 2);
      if(writecnt < 7)
	{
	  steptsp = STEP_SLEEPINUS(tSP30);
	  WDTEnableInus((tSP30 + tSPSU + tCP + tCPSU) * 3 / 2);
	}
    }

  flmcr->PSU = 1;
  sleepinus(tSPSU);
  
  flmcr->P = 1;
  _sleepa(steptsp);

  flmcr->P = 0;
  sleepinus(tCP);

  flmcr->PSU = 0;
  sleepinus(tCPSU);

  WDTDisable();  
  return 0;
}
